"""
Tests confirmation of add/change
on ModelAdmin that utilize caches
and S3 as a storage backend
"""

import os
from importlib import reload

import localstack_client.session
import pkg_resources
import pytest
from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.file_detector import LocalFileDetector

from admin_action_tools.constants import CONFIRM_CHANGE
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.market.admin import shoppingmall_admin
from tests.market.models import Item


class ConfirmWithS3StorageTests(AdminConfirmIntegrationTestCase):
    def setUp(self):
        self.selenium.file_detector = LocalFileDetector()
        session = localstack_client.session.Session(region_name="us-west-1")
        self.s3 = session.resource("s3")
        self.bucket = self.s3.Bucket(settings.AWS_STORAGE_BUCKET_NAME)
        # Delete all current files
        for obj in self.get_all_media():
            obj.delete()
        super().setUp()

    def get_all_media(self):
        return list(self.bucket.objects.filter(Prefix="media/"))

    def tearDown(self):
        reload(shoppingmall_admin)
        # Delete all current files
        for obj in self.get_all_media():
            obj.delete()
        super().tearDown()

    def test_s3_is_being_used(self):
        self.assertTrue(settings.USE_S3)
        self.assertIsNotNone(settings.AWS_ACCESS_KEY_ID)
        self.assertEqual(
            settings.STORAGES["default"]["BACKEND"],
            "tests.storage_backends.PublicMediaStorage",
        )

    def test_should_save_file_additions(self):
        selenium_version = pkg_resources.get_distribution("selenium").parsed_version
        if selenium_version.major < 4:
            pytest.skip("Known issue `https://github.com/SeleniumHQ/selenium/issues/8762` with this selenium version.")

        item = Item.objects.create(name="item", price=1, currency=Item.VALID_CURRENCIES[0][0])

        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)

        # Upload a new file
        self.selenium.find_element(By.ID, "id_file").send_keys(
            os.getcwd() + "/admin_action_tools/tests/snapshot/screenshot.png"
        )

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()
        self.assertRegex(item.file.name, r"screenshot.*\.png$")
        self.assertEqual(21, int(item.price))

        # Check S3 for the file
        objects = self.get_all_media()
        self.assertEqual(len(objects), 1)
        self.assertRegex(objects[0].key, r"screenshot.*\.png$")

    def test_should_save_file_changes(self):
        selenium_version = pkg_resources.get_distribution("selenium").parsed_version
        if selenium_version.major < 4:
            pytest.skip("Known issue `https://github.com/SeleniumHQ/selenium/issues/8762` with this selenium version.")

        file = SimpleUploadedFile(
            name="old_file.jpg",
            content=open("admin_action_tools/tests/snapshot/screenshot.png", "rb").read(),
            content_type="image/jpeg",
        )
        item = Item.objects.create(name="item", price=1, currency=Item.VALID_CURRENCIES[0][0], file=file)

        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)

        # Upload a new file
        self.selenium.find_element(By.ID, "id_file").send_keys(
            os.getcwd() + "/admin_action_tools/tests/snapshot/screenshot.png"
        )

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()
        self.assertEqual(21, int(item.price))
        self.assertRegex(item.file.name, r"screenshot.*\.png$")

        # Check S3 for the file
        objects = self.get_all_media()
        self.assertEqual(len(objects), 2)
        get_last_modified = lambda obj: int(obj.last_modified.strftime("%s"))
        objects_by_last_modified = [obj for obj in sorted(objects, key=get_last_modified)]
        self.assertRegex(objects_by_last_modified[-1].key, r"screenshot.*\.png$")
        self.assertRegex(objects_by_last_modified[0].key, r"old_file.*\.jpg$")

    def test_should_remove_file_if_clear_selected(self):
        file = SimpleUploadedFile(
            name="old_file.jpg",
            content=open("admin_action_tools/tests/snapshot/screenshot.png", "rb").read(),
            content_type="image/jpeg",
        )
        item = Item.objects.create(name="item", price=1, currency=Item.VALID_CURRENCIES[0][0], file=file)

        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)

        # Choose to clear the existing file
        self.selenium.find_element(By.ID, "file-clear_id").click()
        self.assertTrue(self.selenium.find_element(By.XPATH, ".//*[@id='file-clear_id']").get_attribute("checked"))

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()
        self.assertEqual(21, int(item.price))
        # Should have cleared `file` since clear was selected
        self.assertFalse(item.file)

        # Check S3 for the file
        # Since deleting from model instance in Django does not automatically
        #   delete from storage, the old file should still be in S3
        objects = self.get_all_media()
        self.assertEqual(len(objects), 1)
        self.assertRegex(objects[0].key, r"old_file.*\.jpg$")



"""
Ensure that files are saved during confirmation
Without file changes, Django is relied on

With file changes, we cache the object, save it with
the files if new, or add files to existing obj and save

Then send the rest of the changes to Django to handle

This is arguably the most we fiddle with the Django request
Thus we should test it extensively
"""

import time
from unittest import mock

from django.core.cache import cache
from django.core.files.uploadedfile import SimpleUploadedFile

from admin_action_tools.constants import CACHE_KEYS, CONFIRMATION_RECEIVED
from admin_action_tools.file_cache import FileCache
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from admin_action_tools.utils import format_cache_key
from tests.factories import ItemFactory, ShopFactory
from tests.market.admin import ItemAdmin
from tests.market.models import Item, Shop


class TestConfirmationUsingFileCache(AdminConfirmTestCase):
    def setUp(self):
        # Load the Change Item Page
        ItemAdmin.confirm_change = True
        ItemAdmin.fields = ["name", "price", "file", "image", "currency"]
        ItemAdmin.save_as = True
        ItemAdmin.save_as_continue = True

        self.image_path = "admin_action_tools/tests/snapshot/screenshot.png"
        f = SimpleUploadedFile(
            name="test_file.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )
        i = SimpleUploadedFile(
            name="test_image.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )
        self.item = ItemFactory(name="Not name", file=f, image=i)

        return super().setUp()

    def test_save_as_continue_true_should_not_redirect_to_changelist(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = True

        # Upload new image and remove file
        i2 = SimpleUploadedFile(
            name="test_image2.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )
        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_saveasnew": True,
        }

        # Set cache
        cache_item = Item(
            name=data["name"],
            price=data["price"],
            currency=data["currency"],
            image=i2,
        )
        file_cache = FileCache()
        file_cache.set(format_cache_key(model="Item", field="image"), i2)

        cache.set(CACHE_KEYS["object"], cache_item)
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertIn("You may edit it again below.", message)

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/item/{self.item.id + 1}/change/")

        # Should not have changed existing item
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertEqual(item.file.name.count("test_file"), 1)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.file)
        self.assertEqual(new_item.image.name.count("test_image2"), 1)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_save_as_continue_false_should_redirect_to_changelist(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = False

        # Upload new image and remove file
        i2 = SimpleUploadedFile(
            name="test_image2.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )
        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_saveasnew": True,
        }

        # Set cache
        cache_item = Item(
            name=data["name"],
            price=data["price"],
            currency=data["currency"],
            image=i2,
        )
        file_cache = FileCache()
        file_cache.set(format_cache_key(model="Item", field="image"), i2)

        cache.set(CACHE_KEYS["object"], cache_item)
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertEqual(item.file.name.count("test_file"), 1)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.file)
        self.assertEqual(new_item.image.name.count("test_image2"), 1)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_saveasnew_without_any_file_changes_should_save_new_instance_without_files(
        self,
    ):
        item = self.item

        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "image": "",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_saveasnew": True,
        }

        # Set cache
        cache_item = Item(
            name=data["name"],
            price=data["price"],
            currency=data["currency"],
        )

        cache.set(CACHE_KEYS["object"], cache_item)
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertIn("You may edit it again below.", message)

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/item/{self.item.id + 1}/change/")

        # Should not have changed existing item
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertEqual(item.file.name.count("test_file"), 1)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        # In Django (by default), the save as new does not transfer over the files
        self.assertFalse(new_item.file)
        self.assertFalse(new_item.image)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_add_with_upload_file_should_save_new_instance_with_files(self):
        # Upload new file
        f2 = SimpleUploadedFile(
            name="test_file2.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )

        # Request.POST
        data = {
            "name": "name",
            "price": 2.0,
            "image": "",
            "file": f2,
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_add": True,
            "_save": True,
        }

        # Set cache
        cache_item = Item(name=data["name"], price=data["price"], currency=data["currency"], file=f2)

        cache.set(CACHE_KEYS["object"], cache_item)
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_add"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post("/admin/market/item/add/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should not have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        self.item.refresh_from_db()
        self.assertEqual(self.item.name, "Not name")
        self.assertEqual(self.item.file.name.count("test_file"), 1)
        self.assertEqual(self.item.image.name.count("test_image2"), 0)
        self.assertEqual(self.item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=self.item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertIsNotNone(new_item.file)
        self.assertFalse(new_item.image)

        self.assertRegex(new_item.file.name, r"test_file2.*\.jpg$")

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_add_without_cached_post_should_save_new_instance_with_file(self):
        # Upload new file
        f2 = SimpleUploadedFile(
            name="test_file2.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )

        # Request.POST
        data = {
            "name": "name",
            "price": 2.0,
            "image": "",
            "file": f2,
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_add": True,
            "_save": True,
        }

        # Set cache
        cache_item = Item(name=data["name"], price=data["price"], currency=data["currency"], file=f2)

        cache.set(CACHE_KEYS["object"], cache_item)
        # Make sure there's no post cached post
        cache.delete(CACHE_KEYS["post"])

        # Click "Yes, I'm Sure"
        del data["_confirm_add"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post("/admin/market/item/add/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should not have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        self.item.refresh_from_db()
        self.assertEqual(self.item.name, "Not name")
        self.assertEqual(self.item.file.name.count("test_file"), 1)
        self.assertEqual(self.item.image.name.count("test_image2"), 0)
        self.assertEqual(self.item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=self.item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.image)

        # Able to save the cached file since cached object was there even though cached post was not
        self.assertRegex(new_item.file.name, r"test_file2.*\.jpg$")

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_add_without_cached_object_should_save_new_instance_but_not_have_file(self):
        # Request.POST
        data = {
            "name": "name",
            "price": 2.0,
            "image": "",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_add": True,
            "_save": True,
        }

        # Make sure there's no post cached obj
        cache.delete(CACHE_KEYS["object"])
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_add"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post("/admin/market/item/add/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should not have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        self.item.refresh_from_db()
        self.assertEqual(self.item.name, "Not name")
        self.assertEqual(self.item.file.name.count("test_file"), 1)
        self.assertEqual(self.item.image.name.count("test_image2"), 0)
        self.assertEqual(self.item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=self.item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.image)

        # FAILED to save the file, because cached item was not there
        self.assertFalse(new_item.file)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_add_without_any_cache_should_save_new_instance_but_not_have_file(self):
        # Request.POST
        data = {
            "name": "name",
            "price": 2.0,
            "image": "",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_add": True,
            "_save": True,
        }

        # Make sure there's no cache
        cache.delete(CACHE_KEYS["object"])
        cache.delete(CACHE_KEYS["post"])

        # Click "Yes, I'm Sure"
        del data["_confirm_add"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post("/admin/market/item/add/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should not have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        self.item.refresh_from_db()
        self.assertEqual(self.item.name, "Not name")
        self.assertEqual(self.item.file.name.count("test_file"), 1)
        self.assertEqual(self.item.image.name.count("test_image2"), 0)
        self.assertEqual(self.item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=self.item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.image)

        # FAILED to save the file, because cached item was not there
        self.assertFalse(new_item.file)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_change_without_cached_post_should_save_file_changes(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = False

        # Upload new image and remove file
        i2 = SimpleUploadedFile(
            name="test_image2.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )
        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "image": i2,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_saveasnew": True,
        }

        # Set cache
        cache_item = Item(
            name=data["name"],
            price=data["price"],
            currency=data["currency"],
            image=i2,
        )
        file_cache = FileCache()
        file_cache.set(format_cache_key(model="Item", field="image"), i2)

        cache.set(CACHE_KEYS["object"], cache_item)
        # Ensure no cached post
        cache.delete(CACHE_KEYS["post"])

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        # Image would have been in FILES and not in POST
        del data["image"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertEqual(item.file.name.count("test_file"), 1)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        # Should have cleared `file` since clear was selected
        self.assertFalse(new_item.file)
        self.assertIsNotNone(new_item.image)
        # Saved cached file from cached obj even if cached post was missing
        self.assertIn("test_image2", new_item.image.name)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_change_without_cached_object_should_save_but_without_file_changes(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = False

        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_saveasnew": True,
        }

        # Ensure no cached obj
        cache.delete(CACHE_KEYS["object"])
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertEqual(item.file.name.count("test_file"), 1)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.file)
        # FAILED to save image
        self.assertFalse(new_item.image)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_change_without_any_cache_should_save_but_not_have_file_changes(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = False

        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_saveasnew": True,
        }

        # Ensure no cache
        cache.delete(CACHE_KEYS["object"])
        cache.delete(CACHE_KEYS["post"])

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/2/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should not have changed existing item
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertEqual(item.file.name.count("test_file"), 1)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have saved new item
        self.assertEqual(Item.objects.count(), 2)
        new_item = Item.objects.filter(id=item.id + 1).first()
        self.assertIsNotNone(new_item)
        self.assertEqual(new_item.name, data["name"])
        self.assertEqual(new_item.price, data["price"])
        self.assertEqual(new_item.currency, data["currency"])
        self.assertFalse(new_item.file)
        # FAILED to save image
        self.assertFalse(new_item.image)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_change_without_changing_file_should_save_changes(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = False

        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "image": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_save": True,
        }

        # Set cache
        cache_item = Item(
            name=data["name"],
            price=data["price"],
            currency=data["currency"],
        )

        cache.get(CACHE_KEYS["object"], cache_item)
        cache.get(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/1/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should have changed existing item
        self.assertEqual(Item.objects.count(), 1)
        item.refresh_from_db()
        self.assertEqual(item.name, "name")
        # Should have cleared if requested
        self.assertFalse(item.file.name)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    @mock.patch("admin_action_tools.admin.confirm_tool.CACHE_TIMEOUT", 1)
    def test_old_cache_should_not_be_used(self):
        item = self.item

        # Upload new image and remove file
        i2 = SimpleUploadedFile(
            name="test_image2.jpg",
            content=open(self.image_path, "rb").read(),
            content_type="image/jpeg",
        )
        # Click "Save And Continue"
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "image": i2,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(
            rendered_content=response.rendered_content,
            save_action="_continue",
            multipart_form=True,
        )

        # Should have cached the unsaved item
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNotNone(cached_item)

        # Should not have saved the changes yet
        self.assertEqual(Item.objects.count(), 1)
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertIsNotNone(item.file)
        self.assertIsNotNone(item.image)

        # Wait for cache to time out

        time.sleep(1)

        # Check that it did time out
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNone(cached_item)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data["image"] = ""
        data[CONFIRMATION_RECEIVED] = True
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data=data)

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/item/{item.id}/change/")

        # Should have saved item
        self.assertEqual(Item.objects.count(), 1)
        saved_item = Item.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.price, data["price"])
        self.assertEqual(saved_item.currency, data["currency"])
        self.assertFalse(saved_item.file)

        # SHOULD not have saved image since it was in the old cache
        self.assertNotIn("test_image2", saved_item.image)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_cache_with_incorrect_model_should_not_be_used(self):
        item = self.item
        # Load the Change Item Page
        ItemAdmin.save_as_continue = False

        # Request.POST
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_save": True,
        }

        # Set cache to incorrect model
        cache_obj = Shop(name="ShopName")

        cache.set(CACHE_KEYS["object"], cache_obj)
        cache.set(CACHE_KEYS["post"], data)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True

        with mock.patch.object(ItemAdmin, "message_user") as message_user:
            response = self.client.post(f"/admin/market/item/{self.item.id}/change/", data=data)
            # Should show message to user with correct obj and path
            message_user.assert_called_once()
            message = message_user.call_args[0][1]
            self.assertIn("/admin/market/item/1/change/", message)
            self.assertIn(data["name"], message)
            self.assertNotIn("You may edit it again below.", message)

        # Should have redirected to changelist
        self.assertEqual(response.url, "/admin/market/item/")

        # Should have changed existing item
        self.assertEqual(Item.objects.count(), 1)
        item.refresh_from_db()
        self.assertEqual(item.name, "name")
        # Should have cleared if requested
        self.assertFalse(item.file.name)
        self.assertEqual(item.image.name.count("test_image2"), 0)
        self.assertEqual(item.image.name.count("test_image"), 1)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_form_without_files_should_not_use_cache(self):
        cache.delete_many(CACHE_KEYS.values())
        shop = ShopFactory()
        # Click "Save And Continue"
        data = {
            "id": shop.id,
            "name": "name",
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shop/{shop.id}/change/", data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Should not have set cache since not multipart form
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))



import functools
from importlib import import_module
from typing import Callable, Dict, Optional, Type

from django import forms
from django.contrib.admin import helpers
from django.db.models import QuerySet
from django.forms import Form
from django.http import HttpRequest

from admin_action_tools.admin.base import BaseMixin
from admin_action_tools.constants import CONFIRM_FORM, ToolAction
from admin_action_tools.toolchain import ToolChain, add_finishing_step
from admin_action_tools.utils import snake_to_title_case


class ActionFormMixin(BaseMixin):
    # Custom templates (designed to be over-ridden in subclasses)
    action_form_template: str = None

    def build_context(
        self,
        request: HttpRequest,
        func: Callable,
        queryset: QuerySet,
        form_instance: Form,
        tool_name: str,
        display_queryset: bool,
    ):
        action_display_name = snake_to_title_case(func.__name__)
        title = f"Configure Action: {action_display_name}"
        opts = self.model._meta  # pylint: disable=W0212

        return {
            **self.admin_site.each_context(request),
            "title": title,
            "action": func.__name__,
            "action_display_name": action_display_name,
            "action_checkbox_name": helpers.ACTION_CHECKBOX_NAME,
            "submit_name": "confirm_action",
            "queryset": queryset if display_queryset else [],
            "media": self.media + form_instance.media,
            "opts": opts,
            "form": form_instance,
            "submit_action": tool_name,
            "submit_text": "Continue",
            "back_text": "Back",
        }

    def render_action_form(self, request: HttpRequest, context: Dict):
        return super().render_template(
            request, context, "form_tool/action_form.html", custom_template=self.action_form_template
        )

    @staticmethod
    def __get_metadata(form):
        return {
            "type": "form",
            "module": form.__module__,
            "name": form.__name__,
        }

    @staticmethod
    def load_form(data, files, metadata):
        # import_module use sys.module as a caching mechanism
        module = import_module(metadata["module"])
        form = getattr(module, metadata["name"])
        form_instance: Form = form(data, files)
        form_instance.is_valid()
        return form_instance

    def run_form_tool(
        self, func: Callable, request: HttpRequest, queryset_or_object, form: forms, display_queryset: bool
    ):
        tool_chain: ToolChain = ToolChain(request)
        tool_name = f"{CONFIRM_FORM}_{form.__name__}"
        step = tool_chain.get_next_step(tool_name)

        if step == ToolAction.BACK:
            # cancel ask, revert to previous form
            data = tool_chain.rollback()
            form_instance = self.get_instance(form, data=data, instance=queryset_or_object)
        # First called by `Go` which would not have tool_name in params
        elif step == ToolAction.CONFIRMED:
            # form is filled
            form_instance = self.get_instance(
                form, data=request.POST, files=request.FILES, instance=queryset_or_object
            )
            if form_instance.is_valid():
                metadata = self.__get_metadata(form)
                tool_chain.set_tool(tool_name, form_instance.data, form_instance.files, metadata=metadata)
                return func(self, request, queryset_or_object)
        elif step in {ToolAction.FORWARD, ToolAction.CANCEL}:
            # forward to next
            return func(self, request, queryset_or_object)
        else:
            form_instance = self.get_instance(form, instance=queryset_or_object)

        queryset: QuerySet = self.to_queryset(request, queryset_or_object)
        context = self.build_context(request, func, queryset, form_instance, tool_name, display_queryset)

        # Display form
        return self.render_action_form(request, context)

    @staticmethod
    def get_instance(
        form: Type[forms.BaseForm],
        data: Optional[dict] = None,
        files: Optional[dict] = None,
        instance: Optional[forms.BaseForm] = None,
    ) -> forms.BaseForm:
        if issubclass(form, forms.ModelForm):  # pylint: disable=E721
            return form(data, instance=instance, files=files)
        return form(data, files=files)


def add_form_to_action(form: Form, display_queryset=True):
    """
    @add_form_to_action function wrapper for Django ModelAdmin actions
    Will redirect to a form page to ask for more information

    Next, it would call the action with the form data.
    """

    def add_form_to_action_decorator(func):
        # make sure tools chain is setup
        func = add_finishing_step(func)

        @functools.wraps(func)
        def func_wrapper(modeladmin: ActionFormMixin, request, queryset_or_object):
            return modeladmin.run_form_tool(func, request, queryset_or_object, form, display_queryset)

        return func_wrapper

    return add_form_to_action_decorator



from django.http import QueryDict

from admin_action_tools.constants import CONFIRM_FORM
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from admin_action_tools.toolchain import ToolChain
from tests.market.form import NoteActionForm


class TestToolchain(AdminConfirmTestCase):
    def test_toolchain_expired(self):
        request = self.factory.request()
        name = f"toolchain{request.path}"
        request.session[name] = {
            "expire_at": "2012-11-02T15:14:31.000",
            "history": ["tool1"],
            "tool1": {"data": {"field1": True}, "metadata": {}},
        }
        toolchain = ToolChain(request)

        # test toolchain reset
        self.assertEqual(toolchain.get_history(), [])

        # test data is save
        self.assertEqual(request.session[name]["history"], [])

    def test_toolchain_wrong_date(self):
        request = self.factory.request()
        name = f"toolchain{request.path}"
        request.session[name] = {
            "expire_at": "ggg",
            "history": ["tool1"],
            "tool1": {"data": {"field1": True}, "metadata": {}},
        }
        toolchain = ToolChain(request)

        # test toolchain reset
        self.assertEqual(toolchain.get_history(), [])

        # test data is save
        self.assertEqual(request.session[name]["history"], [])

    def test_toolchain_wrong_date_type(self):
        request = self.factory.request()
        name = f"toolchain{request.path}"
        request.session[name] = {
            "expire_at": 3,
            "history": ["tool1"],
            "tool1": {"data": {"field1": True}, "metadata": {}},
        }
        toolchain = ToolChain(request)

        # test toolchain reset
        self.assertEqual(toolchain.get_history(), [])

        # test data is save
        self.assertEqual(request.session[name]["history"], [])

    def test_toolchain_querydict(self):
        data = QueryDict("a=1&a=2&a=3")
        request = self.factory.request()
        toolchain = ToolChain(request)
        res = toolchain._ToolChain__clean_data("Toolname", data, files={}, metadata={})

        self.assertEqual(res["data"], "a=1&a=2&a=3")



from django.urls import reverse
from django.utils import timezone

from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import ShopFactory, TransactionFactory
from tests.market.models import Transaction


class TestModelFieldTypes(AdminConfirmTestCase):
    def test_confirm_add_of_datetime_and_field(self):
        shop = ShopFactory()
        expected_date = timezone.now().date()
        expected_timestamp = timezone.now()
        data = {
            "date": str(expected_date),
            "timestamp_0": str(expected_timestamp.date()),
            "timestamp_1": str(expected_timestamp.time()),
            "currency": "USD",
            "shop": shop.id,
            "total": 0,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_transaction_add"), data)

        # Should not have been added yet
        self.assertEqual(Transaction.objects.count(), 0)

        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/transaction/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertSubmitHtml(rendered_content=response.rendered_content)

        # Confirmation page would not have the _confirm_add sent on submit
        del data["_confirm_add"]
        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        response = self.client.post(reverse("admin:market_transaction_add"), data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/transaction/")
        self.assertEqual(Transaction.objects.count(), 1)

        # Ensure that the date and timestamp saved correctly
        transaction = Transaction.objects.first()
        self.assertEqual(transaction.date, expected_date)
        self.assertEqual(transaction.timestamp, expected_timestamp)

    def test_confirm_change_of_datetime_and_date_field(self):
        transaction = TransactionFactory()
        original_date = transaction.date
        original_timestamp = transaction.timestamp
        data = {
            "id": transaction.id,
            "date": "2021-01-01",
            "timestamp_0": "2021-01-01",
            "timestamp_1": "12:30:00",
            "currency": "USD",
            "shop": transaction.shop.id,
            "total": 0,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/transaction/{transaction.id}/change/", data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/transaction/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Hasn't changed item yet
        transaction.refresh_from_db()
        self.assertEqual(transaction.date, original_date)
        self.assertEqual(transaction.timestamp, original_timestamp)

        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        del data["_confirm_change"]
        response = self.client.post(f"/admin/market/transaction/{transaction.id}/change/", data)
        # will show the change page for this transaction
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f"/admin/market/transaction/{transaction.id}/change/")
        # Should not be the confirmation page, we already confirmed change
        self.assertNotEqual(response.templates, expected_templates)
        self.assertEqual(Transaction.objects.count(), 1)

        transaction.refresh_from_db()
        self.assertEqual(str(transaction.date), "2021-01-01")
        self.assertEqual(str(transaction.timestamp.date()), "2021-01-01")
        self.assertEqual(str(transaction.timestamp.time()), "12:30:00")



"""
Ensures that confirmations work with validators on the Model and on the Modelform.

NOTE: These unit tests are not enough to ensure that confirmations work with validators.
    Please ensure to add integration tests too!
"""

from unittest import mock

from django.urls import reverse
from django.utils import timezone

from admin_action_tools.constants import CONFIRM_ADD
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import (
    InventoryFactory,
    ItemFactory,
    ShopFactory,
    TransactionFactory,
)
from tests.market.models import Checkout, ItemSale


class TestWithValidators(AdminConfirmTestCase):
    @mock.patch("tests.market.models.ItemSale.clean")
    def test_can_confirm_for_models_with_validator_on_model_field(self, _mock_clean):
        # ItemSale.currency has a validator on it
        item = ItemFactory()
        transaction = TransactionFactory()
        data = {
            "transaction": transaction.id,
            "item": item.id,
            "quantity": 1,
            "currency": "USD",
            "total": 10.00,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_itemsale_add"), data)

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)

        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/itemsale/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertSubmitHtml(rendered_content=response.rendered_content)

        # Confirmation page would not have the _confirm_add sent on submit
        del data["_confirm_add"]
        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        response = self.client.post(reverse("admin:market_itemsale_add"), data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/itemsale/")
        self.assertEqual(ItemSale.objects.count(), 1)

        # Ensure that the date and timestamp saved correctly
        item_sale = ItemSale.objects.first()
        self.assertEqual(item_sale.transaction, transaction)
        self.assertEqual(item_sale.item, item)
        self.assertEqual(item_sale.currency, "USD")

    def test_cannot_confirm_for_models_with_validator_on_model_field_if_validator_fails(
        self,
    ):
        # ItemSale.currency has a validator on it
        shop = ShopFactory()
        item = ItemFactory()
        InventoryFactory(shop=shop, item=item, quantity=10)
        transaction = TransactionFactory(shop=shop)
        data = {
            "transaction": transaction.id,
            "item": item.id,
            "quantity": 1,
            "currency": "FAKE",
            "total": 10.00,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_itemsale_add"), data)

        # Should show form with error and not confirmation page
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/itemsale/change_form.html",
            "admin/market/change_form.html",
            "admin/change_form.html",
        ]
        self.assertEqual(response.template_name, expected_templates)
        self.assertEqual(ItemSale.objects.count(), 0)
        self.assertIn("error", str(response.rendered_content))
        self.assertIn("Invalid Currency", str(response.rendered_content))
        # Should still ask for confirmation
        self.assertIn(CONFIRM_ADD, response.rendered_content)

    def test_can_confirm_for_models_with_clean_overridden(self):
        shop = ShopFactory()
        item = ItemFactory()
        InventoryFactory(shop=shop, item=item, quantity=10)
        transaction = TransactionFactory(shop=shop)
        data = {
            "transaction": transaction.id,
            "item": item.id,
            "quantity": 9,
            "currency": "USD",
            "total": 10.00,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_itemsale_add"), data)

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)

        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/itemsale/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertSubmitHtml(rendered_content=response.rendered_content)

        # Confirmation page would not have the _confirm_add sent on submit
        del data["_confirm_add"]
        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        response = self.client.post(reverse("admin:market_itemsale_add"), data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/itemsale/")
        self.assertEqual(ItemSale.objects.count(), 1)

        # Ensure that the date and timestamp saved correctly
        item_sale = ItemSale.objects.first()
        self.assertEqual(item_sale.transaction, transaction)
        self.assertEqual(item_sale.item, item)
        self.assertEqual(item_sale.currency, "USD")

    def test_cannot_confirm_for_models_with_clean_overridden_if_clean_fails(self):
        shop = ShopFactory()
        item = ItemFactory()
        InventoryFactory(shop=shop, item=item, quantity=1)
        transaction = TransactionFactory(shop=shop)
        # Asking to buy more than the shop has in stock
        data = {
            "transaction": transaction.id,
            "item": item.id,
            "quantity": 9,
            "currency": "USD",
            "total": 10.00,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_itemsale_add"), data)

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)

        # Ensure it shows the form and not the confirmation page
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/itemsale/change_form.html",
            "admin/market/change_form.html",
            "admin/change_form.html",
        ]
        self.assertEqual(response.template_name, expected_templates)
        self.assertTrue("error" in str(response.content))
        self.assertTrue("Shop does not have enough of the item stocked" in str(response.content))

        # Should still be asking for confirmation
        self.assertIn(CONFIRM_ADD, response.rendered_content)

        # Fix the issue by buying only what shop has in stock
        data["quantity"] = 1
        # _confirm_add would still be in the POST data
        response = self.client.post(reverse("admin:market_itemsale_add"), data)

        # Should show confirmation page
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/itemsale/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertSubmitHtml(rendered_content=response.rendered_content)

    def test_can_confirm_for_modelform_with_clean_field_and_clean_overridden(self):
        shop = ShopFactory()
        data = {
            "shop": shop.id,
            "currency": "USD",
            "total": 10.00,
            "date": str(timezone.now().date()),
            "timestamp_0": str(timezone.now().date()),
            "timestamp_1": str(timezone.now().time()),
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_checkout_add"), data)

        # Should not have been added yet
        self.assertEqual(Checkout.objects.count(), 0)

        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/checkout/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertSubmitHtml(rendered_content=response.rendered_content)

        # Confirmation page would not have the _confirm_add sent on submit
        del data["_confirm_add"]
        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        response = self.client.post(reverse("admin:market_checkout_add"), data)
        print(response.content)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/checkout/")
        self.assertEqual(Checkout.objects.count(), 1)

        # Ensure that the date and timestamp saved correctly
        checkout = Checkout.objects.first()
        self.assertEqual(checkout.shop, shop)
        self.assertEqual(checkout.total, 10.00)
        self.assertEqual(checkout.currency, "USD")

    def test_cannot_confirm_for_modelform_with_clean_field_overridden_if_validation_fails(
        self,
    ):
        shop = ShopFactory()
        data = {
            "shop": shop.id,
            "currency": "USD",
            "total": "111",
            "date": str(timezone.now().date()),
            "timestamp_0": str(timezone.now().date()),
            "timestamp_1": str(timezone.now().time()),
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_checkout_add"), data)

        # Should not have been added yet
        self.assertEqual(Checkout.objects.count(), 0)

        # Should show form with error and not confirmation page
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/checkout/change_form.html",
            "admin/market/change_form.html",
            "admin/change_form.html",
        ]
        self.assertEqual(response.template_name, expected_templates)
        self.assertEqual(Checkout.objects.count(), 0)
        self.assertIn("error", str(response.rendered_content))
        self.assertIn("Invalid Total 111", str(response.rendered_content))
        # Should still ask for confirmation
        self.assertIn(CONFIRM_ADD, response.rendered_content)

    def test_cannot_confirm_for_modelform_with_clean_overridden_if_validation_fails(
        self,
    ):
        shop = ShopFactory()
        data = {
            "shop": shop.id,
            "currency": "USD",
            "total": "222",
            "date": str(timezone.now().date()),
            "timestamp_0": str(timezone.now().date()),
            "timestamp_1": str(timezone.now().time()),
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_checkout_add"), data)

        # Should not have been added yet
        self.assertEqual(Checkout.objects.count(), 0)

        # Should not have been added yet
        self.assertEqual(Checkout.objects.count(), 0)

        # Should show form with error and not confirmation page
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/checkout/change_form.html",
            "admin/market/change_form.html",
            "admin/change_form.html",
        ]
        self.assertEqual(response.template_name, expected_templates)
        self.assertEqual(Checkout.objects.count(), 0)
        self.assertIn("error", str(response.rendered_content))
        self.assertIn("Invalid Total 222", str(response.rendered_content))
        # Should still ask for confirmation
        self.assertIn(CONFIRM_ADD, response.rendered_content)



import functools
from typing import Callable, Dict

from django.contrib.admin import helpers
from django.contrib.admin.exceptions import DisallowedModelAdminToField
from django.contrib.admin.options import TO_FIELD_VAR
from django.contrib.admin.utils import flatten_fieldsets, unquote
from django.core.cache import cache
from django.core.exceptions import PermissionDenied
from django.db.models import FileField, ImageField, ManyToManyField, Model, QuerySet
from django.forms import ModelForm
from django.http import HttpRequest, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.decorators.cache import cache_control

from admin_action_tools.admin.base import BaseMixin
from admin_action_tools.constants import (
    CACHE_KEYS,
    CACHE_TIMEOUT,
    CONFIRM_ACTION,
    CONFIRM_ADD,
    CONFIRM_CHANGE,
    CONFIRMATION_RECEIVED,
    SAVE,
    SAVE_ACTIONS,
    SAVE_AND_CONTINUE,
    SAVE_AS_NEW,
    ToolAction,
)
from admin_action_tools.templatetags.formatting import back_url
from admin_action_tools.toolchain import ToolChain, add_finishing_step
from admin_action_tools.utils import (
    format_cache_key,
    get_admin_change_url,
    log,
    snake_to_title_case,
)


class AdminConfirmMixin(BaseMixin):
    # Should we ask for confirmation for changes?
    confirm_change = None

    # Should we ask for confirmation for additions?
    confirm_add = None

    # If asking for confirmation, which fields should we confirm for?
    confirmation_fields = None

    # Custom templates (designed to be over-ridden in subclasses)
    change_confirmation_template = None
    action_confirmation_template = None

    def get_confirmation_fields(self, request, obj=None):
        """
        Hook for specifying confirmation fields
        """
        if self.confirmation_fields is not None:
            return self.confirmation_fields

        model_fields = set([field.name for field in self.model._meta.fields])  # pragma: no branch
        admin_fields = set(flatten_fieldsets(self.get_fieldsets(request, obj)))
        return list(model_fields & admin_fields)

    def render_change_confirmation(self, request, context):
        context.update(
            media=self.media,
        )

        return super().render_template(
            request,
            context,
            "confirm_tool/change_confirmation.html",
            custom_template=self.change_confirmation_template,
        )

    def render_action_confirmation(self, request, context):
        opts = self.model._meta

        context.update(
            media=self.media,
            opts=opts,
        )

        return super().render_template(
            request,
            context,
            "confirm_tool/action_confirmation.html",
            custom_template=self.action_confirmation_template,
        )

    @method_decorator(cache_control(private=True))
    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        if request.method == "POST":
            if (not object_id and CONFIRM_ADD in request.POST) or (object_id and CONFIRM_CHANGE in request.POST):
                log("confirmation is asked for")
                self._file_cache.delete_all()
                cache.delete_many(CACHE_KEYS.values())
                return self._change_confirmation_view(request, object_id, form_url, extra_context)
            elif CONFIRMATION_RECEIVED in request.POST:
                return self._confirmation_received_view(request, object_id, form_url, extra_context)
            else:
                self._file_cache.delete_all()
                cache.delete_many(CACHE_KEYS.values())

        extra_context = self._add_confirmation_options_to_extra_context(extra_context)
        return super().changeform_view(request, object_id, form_url, extra_context)

    def _add_confirmation_options_to_extra_context(self, extra_context):
        log(f"Adding confirmation to extra_content {self.confirm_add} {self.confirm_change}")
        return {
            **(extra_context or {}),
            "confirm_add": self.confirm_add,
            "confirm_change": self.confirm_change,
        }

    def _get_changed_data(self, form: ModelForm, model: Model, obj: object, add: bool) -> Dict:
        """
        Given a form, detect the changes on the form from the default values (if add) or
        from the database values of the object (model instance)

        form - Submitted form that is attempting to alter the obj
        model - the model class of the obj
        obj - instance of model which is being altered
        add - are we attempting to add the obj or does it already exist in the database

        Returns a dictionary of the fields and their changed values if any
        """

        def _display_for_changed_data(field, initial_value, new_value):
            if not (isinstance(field, FileField) or isinstance(field, ImageField)):
                return [initial_value, new_value]

            if initial_value:
                if new_value is False:
                    # Clear has been selected
                    return [initial_value.name, None]
                elif new_value:
                    return [initial_value.name, new_value.name]
                else:
                    # No cover: Technically doesn't get called in current code because
                    # This function is only called if there was a difference in the data
                    return [initial_value.name, initial_value.name]  # pragma: no cover

            if new_value:
                return [None, new_value.name]

            return [None, None]

        changed_data = {}
        if add:
            for name, new_value in form.cleaned_data.items():
                # Don't consider default values as changed for adding
                field_object = model._meta.get_field(name)
                default_value = field_object.get_default()
                if new_value is not None and new_value != default_value:
                    # Show what the default value is
                    changed_data[name] = _display_for_changed_data(field_object, default_value, new_value)
        else:
            # Parse the changed data - Note that using form.changed_data would not work because initial is not set
            for name, new_value in form.cleaned_data.items():
                # Since the form considers initial as the value first shown in the form
                # It could be incorrect when user hits save, and then hits "No, go back to edit"
                obj.refresh_from_db()

                field_object = model._meta.get_field(name)
                initial_value = getattr(obj, name)

                # Note: getattr does not work on ManyToManyFields
                if isinstance(field_object, ManyToManyField):
                    initial_value = field_object.value_from_object(obj)

                if initial_value != new_value:
                    changed_data[name] = _display_for_changed_data(field_object, initial_value, new_value)

        return changed_data

    def _confirmation_received_view(self, request, object_id, form_url, extra_context):
        """
        When the form is a multipart form, the object and POST are cached
        This is required because file(s) cannot be programmically uploaded
        ie. There is no way to set a file on the html form

        If the form isn't multipart, this function would not be called.
        If there are no file changes, do nothing to the request and send to Django.

        If there are files uploaded, save the files from cached object to either:
        - the object instance if already exists
        - or save the new object and modify the request from `add` to `change`
        and pass the request to Django
        """
        log("Confirmation has been received")

        def _reconstruct_request_files():
            """
            Reconstruct the file(s) from the file cache (if any).
            Returns a dictionary of field name to cached file
            """
            reconstructed_files = {}

            cached_object = cache.get(CACHE_KEYS["object"])
            # Reconstruct the files from cached object
            if not cached_object:
                log("Warning: no cached_object")
                return

            if not isinstance(cached_object, self.model):
                # Do not use cache if the model doesn't match this model
                log(f"Warning: cached_object is not of type {self.model}")
                return

            query_dict = request.POST

            for field in self.model._meta.get_fields():
                if not (isinstance(field, (FileField, ImageField))):
                    continue

                cached_file = self._file_cache.get(format_cache_key(model=self.model.__name__, field=field.name))

                # If a file was uploaded, the field is omitted from the POST since it's in request.FILES
                if not query_dict.get(field.name):  # pragma: no cover
                    if not cached_file:
                        log(f"Warning: Could not find file cached for field {field.name}")
                    else:
                        reconstructed_files[field.name] = cached_file

            return reconstructed_files

        reconstructed_files = _reconstruct_request_files()
        if reconstructed_files:
            log(f"Found reconstructed files for fields: {reconstructed_files.keys()}")
            obj = None

            # remove the _confirm_add and _confirm_change from post
            modified_post = request.POST.copy()
            if CONFIRM_ADD in modified_post:
                del modified_post[CONFIRM_ADD]  # pragma: no cover
            if CONFIRM_CHANGE in modified_post:
                del modified_post[CONFIRM_CHANGE]  # pragma: no cover

            if object_id and SAVE_AS_NEW not in request.POST:
                # Update the obj with the new uploaded files
                # then pass rest of changes to Django
                obj = self.model.objects.filter(id=object_id).first()
            else:
                # Create the obj and pass the rest as changes to Django
                # (Since we are not handling the formsets/inlines)
                # Note that this results in the "Yes, I'm Sure" submission
                #   act as a `change` not an `add`
                obj = cache.get(CACHE_KEYS["object"])

            # No cover: __reconstruct_request_files currently checks for cached obj so obj won't be None
            if obj:  # pragma: no cover
                for field, file in reconstructed_files.items():
                    log(f"Setting file field {field} to file {file}")
                    setattr(obj, field, file)
                obj.save()
                object_id = str(obj.id)
                # Update the request path, used in the message to user and redirect
                # Used in `self.response_change`
                request.path = get_admin_change_url(obj)

                if SAVE_AS_NEW in request.POST:
                    # We have already saved the new object
                    # So change action to _continue
                    del modified_post[SAVE_AS_NEW]
                    if self.save_as_continue:
                        modified_post[SAVE_AND_CONTINUE] = True
                    else:
                        modified_post[SAVE] = True
                    if "id" in modified_post:
                        del modified_post["id"]
                        modified_post["id"] = object_id

            request.POST = modified_post

        self._file_cache.delete_all()
        cache.delete_many(CACHE_KEYS.values())

        return super()._changeform_view(request, object_id, form_url, extra_context)

    def _get_cleared_fields(self, request):
        """
        Checks for any ImageField or FileField which have been cleared by user.

        Because the form that is generated by Django for the model, would not have the
        `<field>-clear` inputs in them, they have to be injected into the hidden form
        on the confirmation page.
        """
        return [  # pragma: no branch
            input_name.split("-clear")[0] for input_name in request.POST.keys() if input_name.endswith("-clear")
        ]

    def _change_confirmation_view(self, request, object_id, form_url, extra_context):
        # This code is taken from super()._changeform_view
        # https://github.com/django/django/blob/master/django/contrib/admin/options.py#L1575-L1592
        to_field = request.POST.get(TO_FIELD_VAR, request.GET.get(TO_FIELD_VAR))
        if to_field and not self.to_field_allowed(request, to_field):
            raise DisallowedModelAdminToField("The field %s cannot be referenced." % to_field)

        model = self.model
        opts = model._meta

        if SAVE_AS_NEW in request.POST:
            object_id = None

        add = object_id is None
        if add:
            if not self.has_add_permission(request):
                raise PermissionDenied

            obj = None
        else:
            obj = self.get_object(request, unquote(object_id), to_field)
            if obj is None:
                return self._get_obj_does_not_exist_redirect(request, opts, object_id)

            if not self.has_view_or_change_permission(request, obj):
                raise PermissionDenied

        fieldsets = self.get_fieldsets(request, obj)
        ModelForm = self.get_form(request, obj, change=not add, fields=flatten_fieldsets(fieldsets))

        form = ModelForm(request.POST, request.FILES, instance=obj)
        form_validated = form.is_valid()
        if form_validated:
            new_object = self.save_form(request, form, change=not add)
        else:
            new_object = form.instance
        formsets, inline_instances = self._create_formsets(request, new_object, change=not add)
        # End code from super()._changeform_view
        # form.is_valid() checks both errors and "is_bound"
        # If form has errors, show the errors on the form instead of showing confirmation page
        if not form_validated:
            log("Invalid Form: return early")
            log(form.errors)
            # We must ensure that we ask for confirmation when showing errors
            extra_context = self._add_confirmation_options_to_extra_context(extra_context)
            return super()._changeform_view(request, object_id, form_url, extra_context)

        add_or_new = add or SAVE_AS_NEW in request.POST
        # Get changed data to show on confirmation
        changed_data = self._get_changed_data(form, model, obj, add_or_new)

        changed_confirmation_fields = set(self.get_confirmation_fields(request, obj)) & set(changed_data.keys())
        if not bool(changed_confirmation_fields):
            log("No change detected")
            # No confirmation required for changed fields, continue to save
            return super()._changeform_view(request, object_id, form_url, extra_context)

        # Parse the original save action from request
        save_action = None
        # No cover: There would not be a case of not request.POST.keys() and form is valid
        for key in request.POST.keys():  # pragma: no cover
            if key in SAVE_ACTIONS:
                save_action = key
                break

        cleared_fields = []
        if form.is_multipart():
            log("Caching files")
            cache.set(CACHE_KEYS["object"], new_object, CACHE_TIMEOUT)

            # Save files as tempfiles
            for field_name in request.FILES:
                file = request.FILES[field_name]
                self._file_cache.set(format_cache_key(model=model.__name__, field=field_name), file)

            # Handle when files are cleared - since the `form` object would not hold that info
            cleared_fields = self._get_cleared_fields(request)

        log("Render Change Confirmation")
        title_action = _("adding") if add_or_new else _("changing")
        context = {
            **self.admin_site.each_context(request),
            "preserved_filters": self.get_preserved_filters(request),
            "title": f"{_('Confirm')} {title_action} {opts.verbose_name}",
            "object_name": str(obj),
            "object_id": object_id,
            "app_label": opts.app_label,
            "model_name": opts.model_name,
            "opts": opts,
            "obj": obj or new_object,
            "changed_data": changed_data,
            "add": add,
            "save_as_new": SAVE_AS_NEW in request.POST,
            "submit_name": save_action,
            "form": form,
            "cleared_fields": cleared_fields,
            "formsets": formsets,
            **(extra_context or {}),
        }
        return self.render_change_confirmation(request, context)

    def run_confirm_tool(
        self, func: Callable, request: HttpRequest, queryset_or_object, display_form: bool, display_queryset: bool
    ):
        tool_chain: ToolChain = ToolChain(request)
        step = tool_chain.get_next_step(CONFIRM_ACTION)

        # First called by `Go` which would not have confirm_action in params
        if step == ToolAction.CONFIRMED:
            return func(self, request, queryset_or_object)

        if step == ToolAction.CANCEL:
            tool_chain.clear_tool_chain()
            queryset: QuerySet = self.to_queryset(request, queryset_or_object)
            url = back_url(queryset, self.model._meta)
            return HttpResponseRedirect(url)

        form_instance = self.get_tools_result(tool_chain) if display_form else None

        # get_actions will only return the actions that are allowed
        has_perm = self._get_actions(request).get(func.__name__) is not None

        action_display_name = snake_to_title_case(func.__name__)
        title = f"Confirm Action: {action_display_name}"
        queryset: QuerySet = self.to_queryset(request, queryset_or_object)

        context = {
            **self.admin_site.each_context(request),
            "title": title,
            "queryset": queryset if display_queryset else [],
            "has_perm": has_perm,
            "action": func.__name__,
            "action_display_name": action_display_name,
            "action_checkbox_name": helpers.ACTION_CHECKBOX_NAME,
            "submit_name": "confirm_action",
            "submit_action": CONFIRM_ACTION,
            "submit_text": "Confirm",
            "back_text": "Back",
            "forms": form_instance,
            "readonly": True,
        }

        # Display confirmation page
        return self.render_action_confirmation(request, context)


def confirm_action(display_form=True, display_queryset=True):
    """
    @confirm_action() function wrapper for Django ModelAdmin actions
    Will redirect to a confirmation page to ask for confirmation

    Next, it would call the action if confirmed. Otherwise, it would
    return to the changelist without performing action.
    """

    def confirm_action_decorator(func):
        # make sure tools chain is setup
        func = add_finishing_step(func)

        @functools.wraps(func)
        def func_wrapper(modeladmin: AdminConfirmMixin, request, queryset_or_object):
            return modeladmin.run_confirm_tool(func, request, queryset_or_object, display_form, display_queryset)

        return func_wrapper

    return confirm_action_decorator



from unittest import mock

from django import forms
from django.contrib.auth.models import User
from django.urls import reverse

from admin_action_tools.admin.form_tool import ActionFormMixin
from admin_action_tools.constants import CONFIRM_FORM
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import InventoryFactory, ShopFactory
from tests.market.form import NoteActionForm
from tests.market.models import Shop

CONFIRM_FORM_UNIQUE = f"{CONFIRM_FORM}_{NoteActionForm.__name__}"


class TestFormAction(AdminConfirmTestCase):
    @classmethod
    def setUpTestData(cls):
        cls.superuser = User.objects.create_superuser(  # nosec
            username="super", email="super@email.org", password="pass"
        )

    def setUp(self):
        super().setUp()
        self.shop = ShopFactory()
        self.inv = InventoryFactory(shop=self.shop, quantity=10)

    def test_form_action(self):
        response = self.client.post(
            reverse("admin:market_inventory_actions", kwargs={"pk": self.inv.pk, "tool": "add_notes"}),
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmation page, since we clicked Yes, I'm sure
        self.assertEqual(
            response.template_name,
            [
                "admin/market/inventory/form_tool/action_form.html",
                "admin/market/form_tool/action_form.html",
                "admin/form_tool/action_form.html",
            ],
        )

        # The action was to show user a message, and should happen
        self.assertIn("Configure the", response.rendered_content)

    def test_form_action_with_valid_form(self):
        post_params = {CONFIRM_FORM_UNIQUE: ["Continue"], "date_0": "2022-10-11", "date_1": "14:33:21", "note": "Note"}
        response = self.client.post(
            reverse("admin:market_inventory_actions", kwargs={"pk": self.inv.pk, "tool": "add_notes"}),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmation page, since we clicked Yes, I'm sure
        self.assertEqual(response.template_name, "django_object_actions/change_form.html")

        self.inv.refresh_from_db()
        self.assertEqual(self.inv.notes, "This is the default\n\n2022-10-11 14:33:21+00:00\nNote")

    def test_form_action_with_not_valid_form(self):
        post_params = {CONFIRM_FORM_UNIQUE: ["Continue"], "date_0": "", "date_1": "", "note": "Note"}
        response = self.client.post(
            reverse("admin:market_inventory_actions", kwargs={"pk": self.inv.pk, "tool": "add_notes"}),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmation page, since we clicked Yes, I'm sure
        self.assertEqual(
            response.template_name,
            [
                "admin/market/inventory/form_tool/action_form.html",
                "admin/market/form_tool/action_form.html",
                "admin/form_tool/action_form.html",
            ],
        )

        self.assertIn("Configure the", response.rendered_content)
        self.assertIn("This field is required.", response.rendered_content)

    def test_get_instance_with_model_form(self) -> None:
        form_data = mock.MagicMock(spec=dict)
        model_instance = mock.MagicMock()

        model_form_class = forms.modelform_factory(Shop, form=forms.ModelForm, fields="__all__")
        form_instance = ActionFormMixin.get_instance(model_form_class, data=form_data, instance=model_instance)

        self.assertIsInstance(form_instance, model_form_class)
        self.assertEqual(form_instance.data, form_data)
        self.assertEqual(form_instance.instance, model_instance)
        self.assertIsInstance(form_instance, model_form_class)

    def test_get_instance_with_form(self) -> None:
        form_data = mock.MagicMock(spec=dict)

        form_class = forms.Form
        form_instance = ActionFormMixin.get_instance(form_class, data=form_data)

        self.assertIsInstance(form_instance, form_class)
        self.assertEqual(form_instance.data, form_data)
        self.assertFalse(hasattr(form_instance, "instance"))
        self.assertIsInstance(form_instance, form_class)



from unittest import mock

from django.contrib.admin.options import TO_FIELD_VAR
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.http import HttpResponseBadRequest, HttpResponseForbidden
from django.urls import reverse

from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import InventoryFactory, ItemFactory, ShopFactory
from tests.market.admin import InventoryAdmin, ItemAdmin, ShoppingMallAdmin
from tests.market.models import Inventory, Item, ShoppingMall


@mock.patch.object(ShoppingMallAdmin, "inlines", [])
class TestConfirmChangeAndAdd(AdminConfirmTestCase):
    def test_get_add_without_confirm_add(self):
        ItemAdmin.confirm_add = False
        response = self.client.get(reverse("admin:market_item_add"))
        self.assertFalse(response.context_data.get("confirm_add"))
        self.assertNotIn("_confirm_add", response.rendered_content)

    def test_get_add_with_confirm_add(self):
        response = self.client.get(reverse("admin:market_inventory_add"))
        self.assertTrue(response.context_data.get("confirm_add"))
        self.assertIn("_confirm_add", response.rendered_content)

    def test_get_change_without_confirm_change(self):
        response = self.client.get(reverse("admin:market_shop_add"))
        self.assertFalse(response.context_data.get("confirm_change"))
        self.assertNotIn("_confirm_change", response.rendered_content)

    def test_get_change_with_confirm_change(self):
        response = self.client.get(reverse("admin:market_inventory_add"))
        self.assertTrue(response.context_data.get("confirm_change"))
        self.assertIn("_confirm_change", response.rendered_content)

    def test_post_add_without_confirm_add(self):
        data = {"name": "name", "price": 2.0, "currency": Item.VALID_CURRENCIES[0]}
        response = self.client.post(reverse("admin:market_item_add"), data)
        # Redirects to item changelist and item is added
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/item/")
        self.assertEqual(Item.objects.count(), 1)

    def test_post_add_with_confirm_add(self):
        item = ItemFactory()
        shop = ShopFactory()
        data = {
            "shop": shop.id,
            "item": item.id,
            "quantity": 5,
            "_confirm_add": True,
            "_continue": True,
        }
        response = self.client.post(reverse("admin:market_inventory_add"), data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/inventory/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        form_data = {"shop": str(shop.id), "item": str(item.id), "quantity": str(5)}
        self._assertSimpleFieldFormHtml(rendered_content=response.rendered_content, fields=form_data)
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Should not have been added yet
        self.assertEqual(Inventory.objects.count(), 0)

    def test_post_change_with_confirm_change_shoppingmall_name(self):
        # When testing found that even though name was in confirmation_fields
        # When only name changed, `form.is_valid` = False, and thus didn't trigger
        # confirmation page previously, even though it should have

        mall = ShoppingMall.objects.create(name="name")
        data = {
            "id": mall.id,
            "name": "new name",
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
            "_save": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/shoppingmall/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)
        self._assertSubmitHtml(rendered_content=response.rendered_content)

        # Hasn't changed item yet
        mall.refresh_from_db()
        self.assertEqual(mall.name, "name")

    def test_post_change_with_confirm_change(self):
        item = ItemFactory(name="item")
        data = {
            "name": "name",
            "price": 2.0,
            "currency": Item.VALID_CURRENCIES[0],
            "id": item.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
            "_save": True,
        }
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/item/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)
        form_data = {
            "name": "name",
            "price": str(2.0),
            # "id": str(item.id),
            "currency": Item.VALID_CURRENCIES[0][0],
        }

        self._assertSimpleFieldFormHtml(rendered_content=response.rendered_content, fields=form_data)
        self._assertSubmitHtml(rendered_content=response.rendered_content, multipart_form=True)

        # Hasn't changed item yet
        item.refresh_from_db()
        self.assertEqual(item.name, "item")

    def test_post_change_without_confirm_change(self):
        shop = ShopFactory(name="bob")
        data = {"name": "sally"}
        response = self.client.post(f"/admin/market/shop/{shop.id}/change/", data)
        # Redirects to changelist
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/shop/")
        # Shop has changed
        shop.refresh_from_db()
        self.assertEqual(shop.name, "sally")

    def test_get_confirmation_fields_should_default_if_not_set(self):
        expected_fields = [f.name for f in Item._meta.fields if f.name != "id"]
        ItemAdmin.confirmation_fields = None
        ItemAdmin.fields = expected_fields
        admin = ItemAdmin(Item, AdminSite())
        actual_fields = admin.get_confirmation_fields(self.factory.request())
        for field in expected_fields:
            self.assertIn(field, actual_fields)

    def test_get_confirmation_fields_default_should_only_include_fields_shown_on_admin(
        self,
    ):
        admin_fields = ["name", "price"]
        ItemAdmin.confirmation_fields = None
        ItemAdmin.fields = admin_fields
        admin = ItemAdmin(Item, AdminSite())
        actual_fields = admin.get_confirmation_fields(self.factory.request())
        for field in admin_fields:
            self.assertIn(field, actual_fields)

    def test_get_confirmation_fields_if_set(self):
        expected_fields = ["name", "currency"]
        ItemAdmin.confirmation_fields = expected_fields
        admin = ItemAdmin(Item, AdminSite())
        actual_fields = admin.get_confirmation_fields(self.factory.request())
        self.assertEqual(expected_fields, actual_fields)

    def test_custom_template(self):
        expected_template = "market/admin/my_custom_template.html"
        ItemAdmin.change_confirmation_template = expected_template
        admin = ItemAdmin(Item, AdminSite())
        actual_template = admin.render_change_confirmation(self.factory.request(), context={}).template_name
        self.assertEqual(expected_template, actual_template)
        # Clear our setting to not affect other tests
        ItemAdmin.change_confirmation_template = None

    def test_form_invalid(self):
        self.assertEqual(InventoryAdmin.confirmation_fields, ["quantity"])

        inventory = InventoryFactory(quantity=1)
        data = {
            "quantity": 1,
            "shop": "Invalid value",
            "item": "Invalid value",
            "id": inventory.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
        }
        response = self.client.post(f"/admin/market/inventory/{inventory.id}/change/", data)

        # Form invalid should show errors on form
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context_data.get("errors"))
        self.assertEqual(
            response.context_data["errors"][0],
            ["Select a valid choice. That choice is not one of the available choices."],
        )
        # Should not have updated inventory
        inventory.refresh_from_db()
        self.assertEqual(inventory.quantity, 1)

    def test_confirmation_fields_set_with_confirm_change(self):
        self.assertEqual(InventoryAdmin.confirmation_fields, ["quantity"])

        inventory = InventoryFactory()
        another_shop = ShopFactory()
        data = {
            "quantity": inventory.quantity,
            "id": inventory.id,
            "item": inventory.item.id,
            "shop": another_shop.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
        }
        response = self.client.post(f"/admin/market/inventory/{inventory.id}/change/", data)

        # Should not have shown confirmation page since shop did not change
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("admin:market_inventory_changelist"))
        # Should have updated inventory
        inventory.refresh_from_db()
        self.assertEqual(inventory.shop, another_shop)

    def test_confirmation_fields_set_with_confirm_add(self):
        self.assertEqual(InventoryAdmin.confirmation_fields, ["quantity"])

        item = ItemFactory()
        shop = ShopFactory()

        # Don't set quantity - let it default
        data = {"shop": shop.id, "item": item.id, "_confirm_add": True}
        response = self.client.post(reverse("admin:market_inventory_add"), data)
        # No confirmation needed
        self.assertEqual(response.status_code, 302)

        # Should have been added
        self.assertEqual(Inventory.objects.count(), 1)
        new_inventory = Inventory.objects.all().first()
        self.assertEqual(new_inventory.shop, shop)
        self.assertEqual(new_inventory.item, item)
        self.assertEqual(new_inventory.quantity, Inventory._meta.get_field("quantity").default)

    def test_no_change_permissions(self):
        user = User.objects.create_user(username="user", is_staff=True)
        self.client.force_login(user)

        inventory = InventoryFactory()
        data = {
            "quantity": 1000,
            "id": inventory.id,
            "item": inventory.item.id,
            "shop": inventory.shop.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
        }
        response = self.client.post(f"/admin/market/inventory/{inventory.id}/change/", data)

        self.assertEqual(response.status_code, 403)
        self.assertTrue(isinstance(response, HttpResponseForbidden))

        old_quantity = inventory.quantity
        inventory.refresh_from_db()
        self.assertEqual(inventory.quantity, old_quantity)

    def test_no_add_permissions(self):
        user = User.objects.create_user(username="user", is_staff=True)
        self.client.force_login(user)
        item = ItemFactory()
        shop = ShopFactory()
        data = {"shop": shop.id, "item": item.id, "quantity": 5, "_confirm_add": True}
        response = self.client.post(reverse("admin:market_inventory_add"), data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 403)
        self.assertTrue(isinstance(response, HttpResponseForbidden))

        # Should not have been added
        self.assertEqual(Inventory.objects.count(), 0)

    def test_obj_not_found(self):
        inventory = InventoryFactory()
        data = {
            "quantity": 1000,
            "id": 100,
            "item": inventory.item.id,
            "shop": inventory.shop.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
        }
        response = self.client.post("/admin/market/inventory/100/change/", data)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/")
        self.assertEqual(response.reason_phrase, "Found")

        old_quantity = inventory.quantity
        inventory.refresh_from_db()
        self.assertEqual(inventory.quantity, old_quantity)

        self.assertEqual(Inventory.objects.count(), 1)

    def test_handles_to_field_not_allowed(self):
        item = ItemFactory()
        shop = ShopFactory()
        data = {
            "shop": shop.id,
            "item": item.id,
            "quantity": 5,
            "_confirm_add": True,
            TO_FIELD_VAR: "shop",
        }
        response = self.client.post(reverse("admin:market_inventory_add"), data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 400)
        self.assertTrue(isinstance(response, HttpResponseBadRequest))
        self.assertEqual(response.reason_phrase, "Bad Request")

        # Should not have been added
        self.assertEqual(Inventory.objects.count(), 0)



from admin_action_tools.admin.base import color_action
from admin_action_tools.admin.confirm_tool import AdminConfirmMixin, confirm_action
from admin_action_tools.admin.form_tool import ActionFormMixin, add_form_to_action

__all__ = ["AdminConfirmMixin", "confirm_action", "ActionFormMixin", "add_form_to_action", "color_action"]



from django import template
from django.urls import reverse
from django.utils.html import escape
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
def format_change_data_field_value(field_value):
    if isinstance(field_value, str):
        return field_value
    try:
        output = "<ul>"
        for value in iter(field_value):
            output += "<li>" + escape(value) + "</li>"
        output += "</ul>"
        return mark_safe(output)  # nosec
    except Exception:
        return field_value


@register.simple_tag
def verbose_name(obj, fieldname):
    name = obj._meta.get_field(fieldname).verbose_name
    return name.capitalize() if isinstance(name, str) else name


@register.simple_tag
def back_url(queryset, opts):
    if len(queryset) == 1:
        obj = queryset[0]
        return reverse("admin:%s_%s_change" % (opts.app_label, opts.model_name), args=[obj.pk])
    return reverse("admin:%s_%s_changelist" % (opts.app_label, opts.model_name))


@register.simple_tag
def get_file_name(d, key_name):
    return getattr(d.get(key_name, {}), "name", None)



from typing import Dict, List, Optional, Union

from django.contrib.admin.options import IS_POPUP_VAR
from django.db.models import Model, QuerySet
from django.http import HttpRequest
from django.template.response import TemplateResponse

from admin_action_tools.file_cache import FileCache
from admin_action_tools.toolchain import ToolChain


class BaseMixin:
    _file_cache = FileCache()

    actions: Optional[List[str]]

    def get_change_action(self, fieldname):
        actions = getattr(self, fieldname, [])
        change_actions = []
        for name in actions:
            func = getattr(self, name)
            description = self._get_action_description(func, name)
            change_actions.append((func, name, description))
        return change_actions

    def _get_actions(self, request):
        """
        Return a dictionary mapping the names of all actions & object actions for this
        ModelAdmin to a tuple of (callable, name, description) for each action.
        """
        # If self.actions is set to None that means actions are disabled on
        # this page.
        if self.actions is None or IS_POPUP_VAR in request.GET:
            return {}

        actions = self._get_base_actions()
        actions.extend(self.get_change_action("change_actions"))
        actions.extend(self.get_change_action("changelist_actions"))

        actions = self._filter_actions_by_permissions(request, actions)
        return {name: (func, name, desc) for func, name, desc in actions}  # pragma: no branch

    def to_queryset(self, request: HttpRequest, object_or_queryset: Union[QuerySet, Model]) -> QuerySet:
        if not isinstance(object_or_queryset, QuerySet):
            return self.get_queryset(request).filter(pk=object_or_queryset.pk)
        return object_or_queryset

    def render_template(self, request: HttpRequest, context: Dict, template_name: str, custom_template=None):
        opts = self.model._meta
        app_label = opts.app_label

        request.current_app = self.admin_site.name
        tool_chain: ToolChain = ToolChain(request)
        context["first"] = tool_chain.is_first_tool()

        return TemplateResponse(
            request,
            custom_template
            or [
                f"admin/{app_label}/{opts.model_name}/{template_name}",
                f"admin/{app_label}/{template_name}",
                f"admin/{template_name}",
            ],
            context,
        )

    def get_tools_result(self, tool_chain: ToolChain):
        history = tool_chain.get_history()
        forms = []
        for tool_name in history:
            data, files, metadata = tool_chain.get_tool(tool_name)
            forms.append(self.load_form(data, files, metadata))
        return forms


def color_action(color=None, attrs=None):
    """
    @color_action function wrapper for Django ModelAdmin actions
    will color the button
    """

    def add_form_to_action_decorator(func):
        if attrs:
            func.attrs = attrs
        elif color:
            func.attrs = {"style": f"background-color: {color};"}

        return func

    return add_form_to_action_decorator



from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase


class SmokeTest(AdminConfirmIntegrationTestCase):
    def test_load_admin(self):
        self.selenium.get(self.live_server_url + "/admin/")
        self.assertIn("Django", self.selenium.title)
        self.assertIn("Market", self.selenium.page_source)



"""
Tests confirmation of add/change
on ModelAdmin that includes inlines

Does not test confirmation of inline changes
"""

from importlib import reload

import pkg_resources
import pytest
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select

from admin_action_tools.constants import CONFIRM_CHANGE
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.factories import ShopFactory
from tests.market.admin import shoppingmall_admin
from tests.market.models import GeneralManager, ShoppingMall, Town


class ConfirmWithInlinesTests(AdminConfirmIntegrationTestCase):
    def setUp(self):
        self.admin = shoppingmall_admin.ShoppingMallAdmin
        self.admin.inlines = [shoppingmall_admin.ShopInline]
        super().setUp()

    def tearDown(self):
        reload(shoppingmall_admin)
        super().tearDown()

    def test_should_have_hidden_form(self):
        mall = ShoppingMall.objects.create(name="mall")
        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Change name
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated name
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        name = hidden_form.find_element(By.NAME, "name")
        self.assertIn("New Name", name.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist change
        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)

    def test_should_have_hidden_formsets(self):
        # Not having formsets would cause a `ManagementForm tampered with` issue
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory(name=i) for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("This is New Name")

        self.selenium.find_element(By.NAME, "_continue").click()

        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        hidden_form.find_element(By.NAME, "ShoppingMall_shops-TOTAL_FORMS")

        self.selenium.find_element(By.NAME, "_continue").click()

        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)

    def test_should_have_saved_inline_changes(self):
        gm = GeneralManager.objects.create(name="gm")
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)

        shops = [ShopFactory(name=i) for i in range(3)]

        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")

        # Change shops via inline form
        select_shop = Select(self.selenium.find_element(By.NAME, "ShoppingMall_shops-0-shop"))
        select_shop.select_by_value(str(shops[2].id))

        self.selenium.find_element(By.NAME, "_continue").click()

        self.assertIn("Confirm", self.selenium.page_source)

        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        hidden_form.find_element(By.NAME, "ShoppingMall_shops-TOTAL_FORMS")
        self.selenium.find_element(By.NAME, "_continue").click()

        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)
        self.assertIn(shops[2], mall.shops.all())

    def test_should_respect_get_inlines(self):
        # New in Django 3.0
        django_version = pkg_resources.get_distribution("Django").parsed_version
        if django_version.major < 3:
            pytest.skip("get_inlines() introduced in Django 3.0, and is not in this version")

        shoppingmall_admin.ShoppingMallAdmin.inlines = []
        shoppingmall_admin.ShoppingMallAdmin.get_inlines = lambda self, request, obj=None: [
            shoppingmall_admin.ShopInline
        ]

        gm = GeneralManager.objects.create(name="gm")
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)

        shops = [ShopFactory(name=i) for i in range(3)]

        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")

        # Change shops via inline form
        select_shop = Select(self.selenium.find_element(By.NAME, "ShoppingMall_shops-0-shop"))
        select_shop.select_by_value(str(shops[2].id))

        self.selenium.find_element(By.NAME, "_continue").click()

        self.assertIn("Confirm", self.selenium.page_source)

        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        hidden_form.find_element(By.NAME, "ShoppingMall_shops-TOTAL_FORMS")
        self.selenium.find_element(By.NAME, "_continue").click()

        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)
        self.assertIn(shops[2], mall.shops.all())

    def test_should_respect_get_inline_instances(self):
        shoppingmall_admin.ShoppingMallAdmin.inlines = []
        shoppingmall_admin.ShoppingMallAdmin.get_inline_instances = (
            lambda self, request, obj=None: shoppingmall_admin.ShopInline(self.model, self.admin_site)
        )
        gm = GeneralManager.objects.create(name="gm")
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)

        shops = [ShopFactory(name=i) for i in range(3)]

        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")

        # Change shops via inline form
        select_shop = Select(self.selenium.find_element(By.NAME, "ShoppingMall_shops-0-shop"))
        select_shop.select_by_value(str(shops[2].id))

        self.selenium.find_element(By.NAME, "_continue").click()

        self.assertIn("Confirm", self.selenium.page_source)

        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        hidden_form.find_element(By.NAME, "ShoppingMall_shops-TOTAL_FORMS")
        self.selenium.find_element(By.NAME, "_continue").click()

        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)
        self.assertIn(shops[2], mall.shops.all())



from django.urls import reverse

from admin_action_tools.constants import CACHE_KEY_PREFIX, DEBUG


def snake_to_title_case(string: str) -> str:
    return " ".join(string.split("_")).title()


def get_admin_change_url(obj: object) -> str:
    return reverse(
        "admin:%s_%s_change" % (obj._meta.app_label, obj._meta.model_name),
        args=(obj.pk,),
    )


def format_cache_key(model: str, field: str) -> str:
    return f"{CACHE_KEY_PREFIX}__{model}__{field}"


def log(message: str):  # pragma: no cover
    if DEBUG:
        print(message)


def inspect(obj: object):  # pragma: no cover
    if DEBUG:
        print(f"{str(obj): type(obj) - dir(obj)}")



"""
Tests confirmation of add/change
on ModelAdmin that utilize caches
"""

import os
from importlib import reload

import pkg_resources
import pytest
from django.core.files.uploadedfile import SimpleUploadedFile
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.file_detector import LocalFileDetector

from admin_action_tools.constants import CONFIRM_CHANGE
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.market.admin import shoppingmall_admin
from tests.market.models import Item, ShoppingMall


class ConfirmWithInlinesTests(AdminConfirmIntegrationTestCase):
    def setUp(self):
        self.selenium.file_detector = LocalFileDetector()
        super().setUp()

    def tearDown(self):
        reload(shoppingmall_admin)
        super().tearDown()

    def test_models_without_files_should_not_have_confirmation_received(self):
        mall = ShoppingMall.objects.create(name="mall")
        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Change name
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated name
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        name = hidden_form.find_element(By.NAME, "name")
        self.assertIn("New Name", name.get_attribute("value"))

        with self.assertRaises(NoSuchElementException):
            self.selenium.find_element(By.NAME, "_confirmation_received")

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist change
        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)

    def test_models_with_files_should_have_confirmation_received(self):
        item = Item.objects.create(name="item", price=1)
        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Change price
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)
        self.selenium.find_element(By.ID, "id_currency_0").click()

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()

    def test_should_save_file_additions(self):
        selenium_version = pkg_resources.get_distribution("selenium").parsed_version
        if selenium_version.major < 4:
            pytest.skip("Known issue `https://github.com/SeleniumHQ/selenium/issues/8762` with this selenium version.")

        item = Item.objects.create(name="item", price=1, currency=Item.VALID_CURRENCIES[0][0])

        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)

        # Upload a new file
        self.selenium.find_element(By.ID, "id_file").send_keys(
            os.getcwd() + "/admin_action_tools/tests/snapshot/screenshot.png"
        )

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()
        self.assertEqual(21, int(item.price))
        self.assertRegex(item.file.name, r"screenshot.*\.png$")

    def test_should_save_file_changes(self):
        selenium_version = pkg_resources.get_distribution("selenium").parsed_version
        if selenium_version.major < 4:
            pytest.skip("Known issue `https://github.com/SeleniumHQ/selenium/issues/8762` with this selenium version.")

        file = SimpleUploadedFile(
            name="old_file.jpg",
            content=open("admin_action_tools/tests/snapshot/screenshot.png", "rb").read(),
            content_type="image/jpeg",
        )
        item = Item.objects.create(name="item", price=1, currency=Item.VALID_CURRENCIES[0][0], file=file)

        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)

        # Upload a new file
        self.selenium.find_element(By.ID, "id_file").send_keys(
            os.getcwd() + "/admin_action_tools/tests/snapshot/screenshot.png"
        )

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()
        self.assertEqual(21, int(item.price))
        self.assertRegex(item.file.name, r"screenshot.*\.png$")

    def test_should_remove_file_if_clear_selected(self):
        file = SimpleUploadedFile(
            name="old_file.jpg",
            content=open("admin_action_tools/tests/snapshot/screenshot.png", "rb").read(),
            content_type="image/jpeg",
        )
        item = Item.objects.create(name="item", price=1, currency=Item.VALID_CURRENCIES[0][0], file=file)

        self.selenium.get(self.live_server_url + f"/admin/market/item/{item.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys(2)

        # Choose to clear the existing file
        self.selenium.find_element(By.ID, "file-clear_id").click()
        self.assertTrue(self.selenium.find_element(By.XPATH, ".//*[@id='file-clear_id']").get_attribute("checked"))

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated price
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        price = hidden_form.find_element(By.NAME, "price")
        self.assertEqual("21.00", price.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_confirmation_received")
        self.selenium.find_element(By.NAME, "_continue").click()

        item.refresh_from_db()
        self.assertEqual(21, int(item.price))
        # Should have cleared `file` since clear was selected
        self.assertFalse(item.file)



from django.core.files.uploadedfile import SimpleUploadedFile

from admin_action_tools.file_cache import FileCache

file = SimpleUploadedFile(
    name="test_file.jpg",
    content=open("admin_action_tools/tests/snapshot/screenshot.png", "rb").read(),
    content_type="image/jpeg",
)


def test_should_set_file_cache():
    file_cache = FileCache()
    file_cache.set("key", file)
    assert "key" in file_cache.cached_keys  # nosec
    assert file_cache.get("key") is not None  # nosec


def test_should_delete_file_cache():
    file_cache = FileCache()
    file_cache.set("key", file)
    file_cache.delete("key")
    assert "key" not in file_cache.cached_keys  # nosec
    assert file_cache.get("key") is None  # nosec


def test_should_delete_all_file_cache():
    file_cache = FileCache()
    file_cache.set("key", file)
    file_cache.set("key2", file)
    file_cache.delete_all()
    assert len(file_cache.cached_keys) == 0  # nosec
    assert file_cache.get("key") is None  # nosec
    assert file_cache.get("key2") is None  # nosec



import socket

from django.contrib.auth.models import User
from django.core.cache import cache
from django.test import LiveServerTestCase, RequestFactory, TestCase
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.support.ui import Select

from tests.test_project.settings import SELENIUM_HOST


class RequestSessionFactory(RequestFactory):
    def __init__(self, session, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.session = session

    def request(self, **request):
        request = super().request(**request)
        request.session = self.session
        return request


class AdminConfirmTestCase(TestCase):
    """
    Helper TestCase class and common associated assertions
    """

    @classmethod
    def setUpTestData(cls):
        cls.superuser = User.objects.create_superuser(  # nosec
            username="super", email="super@email.org", password="pass"
        )  # nosec

    def setUp(self):
        cache.clear()
        self.client.force_login(self.superuser)
        self.factory = RequestSessionFactory(self.client.session)

    def _assertManyToManyFormHtml(self, rendered_content, options, selected_ids):
        # Form data should be embedded and hidden on confirmation page
        # Should have the correct ManyToMany options selected
        for option in options:
            self.assertIn(
                f'<option value="{option.id}"{" selected" if option.id in selected_ids else ""}>{str(option)}</option>',
                rendered_content,
            )
        # ManyToManyField should be embedded
        self.assertIn("related-widget-wrapper", rendered_content)

    def _assertSubmitHtml(self, rendered_content, save_action="_save", multipart_form=False):
        # Submit should conserve the save action
        self.assertIn(
            f'<input type="submit" value="Yes, I’m sure" name="{save_action}">',
            rendered_content,
        )
        # There should not be _confirm_add or _confirm_change sent in the form on confirmaiton page
        self.assertNotIn("_confirm_add", rendered_content)
        self.assertNotIn("_confirm_change", rendered_content)

        confirmation_received_html = '<input type="hidden" name="_confirmation_received" value="True">'

        if multipart_form:
            # Should have _confirmation_received as a hidden field
            self.assertIn(confirmation_received_html, rendered_content)
        else:
            self.assertNotIn(confirmation_received_html, rendered_content)

    def _assertSimpleFieldFormHtml(self, rendered_content, fields):
        for k, v in fields.items():
            self.assertIn(f'name="{k}"', rendered_content)
            self.assertIn(f'value="{v}"', rendered_content)

    def _assertFormsetsFormHtml(self, rendered_content, inlines):
        for inline in inlines:
            for field in inline.fields:
                self.assertIn("apple", rendered_content)


class AdminConfirmIntegrationTestCase(LiveServerTestCase):
    # Setup/Teardown Methods
    @classmethod
    def setUpClass(cls):
        cls.host = socket.gethostbyname(socket.gethostname())
        cls.selenium = webdriver.Remote(
            command_executor=f"http://{SELENIUM_HOST}:4444/wd/hub", options=webdriver.FirefoxOptions()
        )
        super().setUpClass()

    def setUp(self):
        self.superuser = User.objects.create_superuser(  # nosec
            username="super", email="super@email.org", password="pass"
        )
        self.client.force_login(self.superuser)

        cookie = self.client.cookies["sessionid"]
        self.selenium.get(
            self.live_server_url + "/admin/"
        )  # selenium will set cookie domain based on current page domain
        self.selenium.add_cookie({"name": "sessionid", "value": cookie.value, "secure": False, "path": "/"})
        return super().setUp()

    def tearDown(self):
        cache.clear()
        self.selenium.get(self.live_server_url)
        return super().tearDown()

    @classmethod
    def tearDownClass(cls):
        cls.selenium.quit()
        super().tearDownClass()

    # Helper Methods

    def set_value(self, by, by_value, value):
        element = self.selenium.find_element(by, by_value)
        element.clear()
        element.send_keys(value)

    def select_value(self, by, by_value, value):
        element = Select(self.selenium.find_element(by, by_value))
        element.select_by_value(value)

    def select_index(self, by, by_value, index):
        element = Select(self.selenium.find_element(by, by_value))
        element.select_by_index(index)

    def print_hidden_form(self):
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        print(hidden_form.get_attribute("innerHTML"))
